# -*- coding: utf-8 -*-

"""
Improve the original dataclasses module.
"""

import typing as T
import enum
import dataclasses

__version__ = "0.1.1"


class MetadataKeyEnum(str, enum.Enum):
    CONVERTER = "_better_dataclass_converter"


T_DATA = T.Dict[str, T.Any]
T_FIELDS = T.Dict[str, dataclasses.Field]

_class_fields: T.Dict[T.Any, T_FIELDS] = {}

T_DATA_LIKE = T.Union[T_DATA, "T_DATA_CLASS", None]


class DataClass:
    """
    Enhanced dataclass that does serialization and deserialization correctly,
    even if the dataclass has nested dataclass fields.

    Usage example::

        @dataclasses.dataclass
        class Profile(DataClass):
            firstname: str = dataclasses.field()
            lastname: str = dataclasses.field()
            ssn: str = dataclasses.field()

        @dataclasses.dataclass
        class Degree(DataClass):
            name: str = dataclasses.field()
            year: int = dataclasses.field()

        @dataclasses.dataclass
        class People(DataClass):
            id: int = dataclasses.field()
            profile: T.Optional[Profile] = Profile.nested_field(default=None)
            degrees: T.Optional[T.List[Degree]] = Degree.list_of_nested_field(default_factory=list)

        people = People(
            id=1,
            profile=Profile(
                firstname="David",
                lastname="John",
                ssn="123-45-6789",
            ),
            degrees=[
                Degree(name="Bachelor", year=2004),
                Degree(name="Master", year=2006),
            ],
        )
        people_data = people.to_dict()
        people1 = People.from_dict(people_data)
    """

    @classmethod
    def get_fields(cls) -> T_FIELDS:
        """
        Get the dict view of the ``dataclasses.Field`` in this class.
        It leverages the cache to avoid the overhead of ``dataclasses.fields``
        function call.
        """
        try:
            return _class_fields[cls]
        except KeyError:
            _class_fields[cls] = {
                field.name: field for field in dataclasses.fields(cls)
            }
            return _class_fields[cls]

    def to_dict(self) -> T_DATA:
        """
        Serialize the dataclass instance to a dict.
        """
        return dataclasses.asdict(self)

    @classmethod
    def from_dict(
        cls: T.Type["T_DATA_CLASS"],
        dct_or_obj: T_DATA_LIKE,
    ) -> T.Optional["DataClass"]:
        """
        Construct an instance from dataclass-like data.
        It could be a dictionary, an instance of this class, or None.
        """
        if isinstance(dct_or_obj, dict):
            _fields = cls.get_fields()
            kwargs = {}
            for k, v in dct_or_obj.items():
                field = _fields[k]
                if MetadataKeyEnum.CONVERTER.value in field.metadata:
                    kwargs[k] = field.metadata[MetadataKeyEnum.CONVERTER](v)
                else:
                    kwargs[k] = v
            return cls(**kwargs)
        elif isinstance(dct_or_obj, cls):
            return dct_or_obj
        elif dct_or_obj is None:
            return None
        else:  # pragma: no cover
            raise TypeError

    @classmethod
    def from_list(
        cls: T.Type["T_DATA_CLASS"],
        list_of_dct_or_obj: T.Optional[T.List[T_DATA_LIKE]],
    ) -> T.Optional[T.List[T.Optional["T_DATA_CLASS"]]]:
        """
        Construct list of instance from list of dataclass-like data.
        It could be a dictionary, an instance of this class, or None.
        """
        if isinstance(list_of_dct_or_obj, list):
            return [cls.from_dict(item) for item in list_of_dct_or_obj]
        elif list_of_dct_or_obj is None:
            return None
        else:  # pragma: no cover
            raise TypeError

    @classmethod
    def _from_mapper(
        cls: T.Type["T_DATA_CLASS"],
        map_of_dct_or_obj: T.Optional[T.Dict[str, T_DATA_LIKE]],
    ) -> T.Optional[T.Dict[str, T.Optional["T_DATA_CLASS"]]]:
        """
        Construct dict of instance from dict of dataclass-like data.
        It could be a dictionary, an instance of this class, or None.
        """
        if isinstance(map_of_dct_or_obj, dict):
            return {k: cls.from_dict(v) for k, v in map_of_dct_or_obj.items()}
        elif map_of_dct_or_obj is None:
            return None
        else:  # pragma: no cover
            raise TypeError

    @classmethod
    def nested_field(
        cls,
        default=dataclasses.MISSING,
        default_factory=dataclasses.MISSING,
        init=True,
        repr=True,
        hash=None,
        compare=True,
        metadata=None,
        **kwargs,
    ):
        """
        Declare a field that is another dataclass.
        """
        if metadata is None:
            metadata = {}
        metadata[MetadataKeyEnum.CONVERTER.value] = cls.from_dict
        params = dict(
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=metadata,
        )
        if default is not dataclasses.MISSING:
            params["default"] = default
        if default_factory is not dataclasses.MISSING:
            params["default_factory"] = default_factory
        params.update(kwargs)
        return dataclasses.field(**params)

    @classmethod
    def list_of_nested_field(
        cls,
        default=dataclasses.MISSING,
        default_factory=dataclasses.MISSING,
        init=True,
        repr=True,
        hash=None,
        compare=True,
        metadata=None,
        **kwargs,
    ):
        """
        Declare a field that is a list of other dataclass.
        """
        if metadata is None:
            metadata = {}
        metadata[MetadataKeyEnum.CONVERTER.value] = cls.from_list
        params = dict(
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=metadata,
        )
        if default is not dataclasses.MISSING:
            params["default"] = default
        if default_factory is not dataclasses.MISSING:
            params["default_factory"] = default_factory
        params.update(kwargs)
        return dataclasses.field(**params)

    @classmethod
    def map_of_nested_field(
        cls,
        default=dataclasses.MISSING,
        default_factory=dataclasses.MISSING,
        init=True,
        repr=True,
        hash=None,
        compare=True,
        metadata=None,
        **kwargs,
    ):
        """
        Declare a field that is a list of other dataclass.
        """
        if metadata is None:
            metadata = {}
        metadata[MetadataKeyEnum.CONVERTER.value] = cls._from_mapper
        params = dict(
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=metadata,
        )
        if default is not dataclasses.MISSING:
            params["default"] = default
        if default_factory is not dataclasses.MISSING:
            params["default_factory"] = default_factory
        params.update(kwargs)
        return dataclasses.field(**params)


T_DATA_CLASS = T.TypeVar("T_DATA_CLASS", bound=DataClass)



# -*- coding: utf-8 -*-

"""
Module for type hints.

See https://github.com/pola-rs/polars/blob/main/py-polars/polars/_typing.py
"""

try:
    import typing_extensions as T
except ImportError:  # pragma: no cover
    import typing as T

if T.TYPE_CHECKING:
    from datetime import date, datetime, time, timedelta
    from decimal import Decimal

    from .expr.api import T_EXPR
    from .dfop.api import T_DFOP

# fmt: off

NumericLiteral: T.TypeAlias = T.Union[int, float, "Decimal"]
TemporalLiteral: T.TypeAlias = T.Union["date", "time", "datetime", "timedelta"]
NonNestedLiteral: T.TypeAlias = T.Union[NumericLiteral, TemporalLiteral, str, bool, bytes]
# Python literal types (can convert into a `lit` expression)
PythonLiteral: T.TypeAlias = T.Union[NonNestedLiteral, T.List[T.Any]]
# Inputs that can convert into a `col` expression
IntoExprColumn: T.TypeAlias = T.Union["T_EXPR", str]
# Inputs that can convert into an expression
IntoExpr: T.TypeAlias = T.Union[PythonLiteral, "T_EXPR", None]
# Inputs that can be interactive with +, -, *, /, etc.
OtherExpr: T.TypeAlias = T.Union[PythonLiteral, "T_EXPR"]

# selector type, and related collection/sequence
ColumnNameOrSelector: T.TypeAlias = T.Union[str, "T_EXPR"]

# fmt: on



# -*- coding: utf-8 -*-

from .base_expr import ExprEnum
from .base_expr import parse_expr
from .expr.api import T_EXPR
from .expr import api as expr
from .base_dfop import DfopEnum
from .base_dfop import parse_dfop
from .dfop.api import T_DFOP
from .dfop import api as dfop



# -*- coding: utf-8 -*-

import typing as T
import enum
import dataclasses

import polars as pl
from .vendor.better_dataclasses import DataClass

from .sentinel import NOTHING, REQUIRED, OPTIONAL

if T.TYPE_CHECKING:  # pragma: no cover
    from .expr.api import T_EXPR


class ExprEnum(str, enum.Enum):
    # Aggregation
    agg = "agg"
    agg_groups = "agg_groups"
    arg_max = "arg_max"
    arg_min = "arg_min"
    agg_count = "agg_count"
    agg_first = "agg_first"
    agg_implode = "agg_implode"
    agg_last = "agg_last"
    agg_len = "agg_len"
    agg_max = "agg_max"
    agg_mean = "agg_mean"
    agg_median = "agg_median"
    agg_min = "agg_min"
    agg_nan_max = "agg_nan_max"
    agg_nan_min = "agg_nan_min"
    agg_product = "agg_product"
    agg_quantile = "agg_quantile"
    agg_std = "agg_std"
    agg_sum = "agg_sum"
    agg_var = "agg_var"
    # Array
    arr = "arr"
    arr_max = "arr_max"
    arr_min = "arr_min"
    arr_median = "arr_median"
    arr_sum = "arr_sum"
    arr_std = "arr_std"
    arr_to_list = "arr_to_list"
    arr_unique = "arr_unique"
    arr_n_unique = "arr_n_unique"
    arr_var = "arr_var"
    arr_all = "arr_all"
    arr_any = "arr_any"
    arr_sort = "arr_sort"
    arr_reverse = "arr_reverse"
    arr_arg_min = "arr_arg_min"
    arr_arg_max = "arr_arg_max"
    arr_get = "arr_get"
    arr_first = "arr_first"
    arr_last = "arr_last"
    arr_join = "arr_join"
    arr_explode = "arr_explode"
    arr_contains = "arr_contains"
    arr_count_matches = "arr_count_matches"
    arr_to_struct = "arr_to_struct"
    arr_shift = "arr_shift"
    # Binary
    binary = "binary"
    binary_contains = "binary_contains"
    binary_decode = "binary_decode"
    binary_encode = "binary_encode"
    binary_ends_with = "binary_ends_with"
    binary_size = "binary_size"
    binary_starts_with = "binary_starts_with"
    # Boolean
    # Categories
    # Columns / names
    column = "column"
    alias = "alias"
    # Computation
    # Functions
    lit = "lit"
    plus = "plus"
    minus = "minus"
    multiple = "multiple"
    divide = "divide"
    # List
    list = "list"
    list_all = "list_all"
    list_any = "list_any"
    list_drop_nulls = "list_drop_nulls"
    list_arg_max = "list_arg_max"
    list_arg_min = "list_arg_min"
    list_concat = "list_concat"
    list_contains = "list_contains"
    list_count_matches = "list_count_matches"
    list_diff = "list_diff"
    list_eval = "list_eval"
    list_explode = "list_explode"
    list_first = "list_first"
    list_gather = "list_gather"
    list_get = "list_get"
    list_head = "list_head"
    list_join = "list_join"
    list_last = "list_last"
    list_len = "list_len"
    list_max = "list_max"
    list_mean = "list_mean"
    list_median = "list_median"
    list_min = "list_min"
    list_reverse = "list_reverse"
    list_sample = "list_sample"
    list_set_difference = "list_set_difference"
    list_set_intersection = "list_set_intersection"
    list_set_symmetric_difference = "list_set_symmetric_difference"
    list_set_union = "list_set_union"
    list_shift = "list_shift"
    list_slice = "list_slice"
    list_sort = "list_sort"
    list_std = "list_std"
    list_sum = "list_sum"
    list_tail = "list_tail"
    list_to_array = "list_to_array"
    list_to_struct = "list_to_struct"
    list_unique = "list_unique"
    list_n_unique = "list_n_unique"
    list_var = "list_var"
    list_gather_every = "list_gather_every"
    # Manipulation / selection
    append = "append"
    arg_sort = "arg_sort"
    arg_true = "arg_true"
    backward_fill = "backward_fill"
    bottom_k = "bottom_k"
    bottom_k_by = "bottom_k_by"
    cast = "cast"
    ceil = "ceil"
    clip = "clip"
    cut = "cut"
    drop_nans = "drop_nans"
    drop_nulls = "drop_nulls"
    explode = "explode"
    extend_constant = "extend_constant"
    fill_nan = "fill_nan"
    fill_null = "fill_null"
    filter = "filter"
    flatten = "flatten"
    floor = "floor"
    forward_fill = "forward_fill"
    gather = "gather"
    gather_every = "gather_every"
    get = "get"
    head = "head"
    inspect = "inspect"
    interpolate = "interpolate"
    interpolate_by = "interpolate_by"
    limit = "limit"
    lower_bound = "lower_bound"
    pipe = "pipe"
    qcut = "qcut"
    rechunk = "rechunk"
    reinterpret = "reinterpret"
    repeat_by = "repeat_by"
    replace = "replace"
    replace_strict = "replace_strict"
    reshape = "reshape"
    reverse = "reverse"
    rle = "rle"
    rle_id = "rle_id"
    round = "round"
    round_sig_figs = "round_sig_figs"
    sample = "sample"
    shift = "shift"
    shrink_dtype = "shrink_dtype"
    shuffle = "shuffle"
    slice = "slice"
    sort = "sort"
    sort_by = "sort_by"
    tail = "tail"
    to_physical = "to_physical"
    top_k = "top_k"
    top_k_by = "top_k_by"
    upper_bound = "upper_bound"
    where = "where"
    # Meta
    # Miscellaneous
    # Name
    # Operators
    and_ = "and"
    or_ = "or"
    eq = "eq"
    eq_missing = "eq_missing"
    ge = "ge"
    gt = "gt"
    le = "le"
    lt = "lt"
    ne = "ne"
    ne_missing = "ne_missing"
    add = "add"
    floordiv = "floordiv"
    mod = "mod"
    mul = "mul"
    neg = "neg"
    sub = "sub"
    truediv = "truediv"
    pow = "pow"
    xor = "xor"
    # String
    string = "string"
    str_concat = "str_concat"
    str_contains = "str_contains"
    str_contains_any = "str_contains_any"
    str_count_matches = "str_count_matches"
    str_decode = "str_decode"
    str_encode = "str_encode"
    str_ends_with = "str_ends_with"
    str_explode = "str_explode"
    str_extract = "str_extract"
    str_extract_all = "str_extract_all"
    str_extract_groups = "str_extract_groups"
    str_extract_many = "str_extract_many"
    str_find = "str_find"
    str_head = "str_head"
    str_join = "str_join"
    str_json_decode = "str_json_decode"
    str_json_path_match = "str_json_path_match"
    str_len_bytes = "str_len_bytes"
    str_len_chars = "str_len_chars"
    str_pad_end = "str_pad_end"
    str_pad_start = "str_pad_start"
    str_replace = "str_replace"
    str_replace_all = "str_replace_all"
    str_replace_many = "str_replace_many"
    str_reverse = "str_reverse"
    str_slice = "str_slice"
    str_split = "str_split"
    str_split_exact = "str_split_exact"
    str_splitn = "str_splitn"
    str_starts_with = "str_starts_with"
    str_strip_chars = "str_strip_chars"
    str_strip_chars_start = "str_strip_chars_start"
    str_strip_chars_end = "str_strip_chars_end"
    str_strip_prefix = "str_strip_prefix"
    str_strip_suffix = "str_strip_suffix"
    str_strptime = "str_strptime"
    str_tail = "str_tail"
    str_to_date = "str_to_date"
    str_to_datetime = "str_to_datetime"
    str_to_decimal = "str_to_decimal"
    str_to_integer = "str_to_integer"
    str_to_lowercase = "str_to_lowercase"
    str_to_titlecase = "str_to_titlecase"
    str_to_time = "str_to_time"
    str_to_uppercase = "str_to_uppercase"
    str_zfill = "str_zfill"
    # Struct
    struct = "struct"
    struct_field = "struct_field"
    struct_json_encode = "struct_json_encode"
    struct_rename_fields = "struct_rename_fields"
    struct_with_fields = "struct_with_fields"
    # Temporal
    dt = "datetime"
    dt_add_business_days = "dt_add_business_days"
    dt_base_utc_offset = "dt_base_utc_offset"
    dt_cast_time_unit = "dt_cast_time_unit"
    dt_century = "dt_century"
    dt_combine = "dt_combine"
    dt_convert_time_zone = "dt_convert_time_zone"
    dt_date = "dt_date"
    dt_datetime = "dt_datetime"
    dt_day = "dt_day"
    dt_dst_offset = "dt_dst_offset"
    dt_epoch = "dt_epoch"
    dt_hour = "dt_hour"
    dt_is_leap_year = "dt_is_leap_year"
    dt_iso_year = "dt_iso_year"
    dt_microsecond = "dt_microsecond"
    dt_millennium = "dt_millennium"
    dt_millisecond = "dt_millisecond"
    dt_minute = "dt_minute"
    dt_month = "dt_month"
    dt_month_end = "dt_month_end"
    dt_month_start = "dt_month_start"
    dt_nanosecond = "dt_nanosecond"
    dt_offset_by = "dt_offset_by"
    dt_ordinal_day = "dt_ordinal_day"
    dt_quarter = "dt_quarter"
    dt_replace_time_zone = "dt_replace_time_zone"
    dt_round = "dt_round"
    dt_second = "dt_second"
    dt_strftime = "dt_strftime"
    dt_time = "dt_time"
    dt_timestamp = "dt_timestamp"
    dt_to_string = "dt_to_string"
    dt_total_days = "dt_total_days"
    dt_total_hours = "dt_total_hours"
    dt_total_microseconds = "dt_total_microseconds"
    dt_total_milliseconds = "dt_total_milliseconds"
    dt_total_minutes = "dt_total_minutes"
    dt_total_nanoseconds = "dt_total_nanoseconds"
    dt_total_seconds = "dt_total_seconds"
    dt_truncate = "dt_truncate"
    dt_week = "dt_week"
    dt_weekday = "dt_weekday"
    dt_with_time_unit = "dt_with_time_unit"
    dt_year = "dt_year"
    # Window


@dataclasses.dataclass
class BaseExpr(DataClass):
    type: str = dataclasses.field(default=REQUIRED)

    def _validate(self):
        for k, v in dataclasses.asdict(self).items():
            if v is REQUIRED:  # pragma: no cover
                raise ValueError(f"Field {k!r} is required for {self.__class__}.")

    def __post_init__(self):
        self._validate()

    def to_polars(self) -> pl.Expr:
        raise NotImplementedError()


expr_enum_to_klass_mapping: T.Dict[str, T.Type["T_EXPR"]] = dict()


def parse_expr(dct: T.Dict[str, T.Any]) -> "T_EXPR":
    """
    Note: you have to import everything in the :mod:`jsonpolars.expr` module
    to make this work.
    """
    return expr_enum_to_klass_mapping[dct["type"]].from_dict(dct)



# -*- coding: utf-8 -*-

"""
Base DataFrame operation.
"""

import typing as T
import enum
import dataclasses

from .vendor.better_dataclasses import DataClass

from .sentinel import NOTHING, REQUIRED, OPTIONAL

if T.TYPE_CHECKING:  # pragma: no cover
    from .dfop.api import T_DFOP


class DfopEnum(str, enum.Enum):
    """ """

    # Aggregation
    count = "count"
    max = "max"
    max_horizontal = "max_horizontal"
    mean = "mean"
    mean_horizontal = "mean_horizontal"
    median = "median"
    min = "min"
    min_horizontal = "min_horizontal"
    product = "product"
    quantile = "quantile"
    std = "std"
    sum = "sum"
    sum_horizontal = "sum_horizontal"
    var = "var"
    # Attributes
    # Computation
    # Descriptive
    # Export
    # GroupBy
    group_by_agg = "group_by_agg"
    group_by_all = "group_by_all"
    group_by_count = "group_by_count"
    group_by_first = "group_by_first"
    group_by_head = "group_by_head"
    group_by_last = "group_by_last"
    group_by_len = "group_by_len"
    group_by_map_groups = "group_by_map_groups"
    group_by_max = "group_by_max"
    group_by_mean = "group_by_mean"
    group_by_median = "group_by_median"
    group_by_min = "group_by_min"
    group_by_n_unique = "group_by_n_unique"
    group_by_quantile = "group_by_quantile"
    group_by_sum = "group_by_sum"
    group_by_tail = "group_by_tail"
    # Manipulation / selection
    bottom_k = "bottom_k"
    cast = "cast"
    clear = "clear"
    clone = "clone"
    drop = "drop"
    drop_in_place = "drop_in_place"
    drop_nulls = "drop_nulls"
    explode = "explode"
    extend = "extend"
    fill_nan = "fill_nan"
    fill_null = "fill_null"
    filter = "filter"
    gather_every = "gather_every"
    get_column = "get_column"
    get_column_index = "get_column_index"
    get_columns = "get_columns"
    group_by = "group_by"
    group_by_dynamic = "group_by_dynamic"
    head = "head"
    hstack = "hstack"
    insert_column = "insert_column"
    interpolate = "interpolate"
    item = "item"
    iter_columns = "iter_columns"
    iter_rows = "iter_rows"
    iter_slices = "iter_slices"
    join = "join"
    join_asof = "join_asof"
    limit = "limit"
    melt = "melt"
    merge_sorted = "merge_sorted"
    partition_by = "partition_by"
    pipe = "pipe"
    pivot = "pivot"
    rechunk = "rechunk"
    rename = "rename"
    replace_column = "replace_column"
    reverse = "reverse"
    rolling = "rolling"
    row = "row"
    rows = "rows"
    rows_by_key = "rows_by_key"
    sample = "sample"
    select = "select"
    select_seq = "select_seq"
    set_sorted = "set_sorted"
    shift = "shift"
    shrink_to_fit = "shrink_to_fit"
    slice = "slice"
    sort = "sort"
    sql = "sql"
    tail = "tail"
    to_dummies = "to_dummies"
    to_series = "to_series"
    top_k = "top_k"
    transpose = "transpose"
    unique = "unique"
    unnest = "unnest"
    unpivot = "unpivot"
    unstack = "unstack"
    update = "update"
    upsample = "upsample"
    vstack = "vstack"
    with_columns = "with_columns"
    with_columns_seq = "with_columns_seq"
    with_row_count = "with_row_count"
    with_row_index = "with_row_index"
    # Miscellaneous
    # Plot
    # Style


@dataclasses.dataclass
class BaseDfop(DataClass):
    type: str = dataclasses.field(default=REQUIRED)

    def _validate(self):
        for k, v in dataclasses.asdict(self).items():
            if v is REQUIRED:  # pragma: no cover
                raise ValueError(f"Field {k!r} is required for {self.__class__}.")

    def __post_init__(self):
        self._validate()


dfop_enum_to_klass_mapping: T.Dict[str, T.Type["T_DFOP"]] = dict()


def parse_dfop(dct: T.Dict[str, T.Any]) -> "T_DFOP":
    """
    Note: you have to import everything in the :mod:`jsonpolars.dfop` module
    to make this work.
    """
    return dfop_enum_to_klass_mapping[dct["type"]].from_dict(dct)



# -*- coding: utf-8 -*-

from datetime import datetime, timedelta, timezone

import polars as pl

from jsonpolars.expr import api as expr
from jsonpolars.tests.expr_case import Case

# fmt: off
case_datetime_to_string_1 = Case(
    input_records=[
        {"time": datetime(2024, 8, 1, 12, 30, 45)},
    ],
    expr=expr.DtToString(
        expr=expr.Datetime(expr=expr.Column(name="time")),
        format="%m/%d/%Y %H:%M:%S",
    ),
    expected_output_records=[
        {"time": "08/01/2024 12:30:45"},
    ],
)
case_datetime_to_string_2 = Case(
    input_records=[
        {"time": datetime(2024, 8, 1, 12, 30, 45)},
    ],
    expr=expr.DtToString(
        expr=expr.Column(name="time"),
        format="%m/%d/%Y %H:%M:%S",
    ),
    expected_output_records=[
        {"time": "08/01/2024 12:30:45"},
    ],
)
case_dt_year = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtYear(
        expr=expr.Column(name="dt"),
    ),
    expected_output_records=[
        {"dt": 2021},
    ],
)
case_dt_quarter = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtQuarter(
        expr=expr.Column(name="dt"),
    ),
    expected_output_records=[
        {"dt": 2},
    ],
)
case_dt_month = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtMonth(
        expr=expr.Column(name="dt"),
    ),
    expected_output_records=[
        {"dt": 6},
    ],
)
case_dt_day = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtDay(
        expr=expr.Column(name="dt"),
    ),
    expected_output_records=[
        {"dt": 15},
    ],
)
case_dt_hour = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtHour(
        expr=expr.Column(name="dt"),
    ),
    expected_output_records=[
        {"dt": 14},
    ],
)
case_dt_minute = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtMinute(
        expr=expr.Column(name="dt"),
    ),
    expected_output_records=[
        {"dt": 30},
    ],
)
case_dt_second = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123)},
    ],
    expr=expr.DtSecond(
        expr=expr.Column(name="dt"),
    ),
    expected_output_records=[
        {"dt": 45},
    ],
)
case_dt_nanosecond = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtNanoSecond(
        expr=expr.Column(name="dt"),
    ),
    expected_output_records=[
        {"dt": 123000000},
    ],
)
case_dt_epoch = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtEpoch(
        expr=expr.Column(name="dt"),
    ),
    expected_output_records=[
        {"dt": 1623767445123000},
    ],
)
case_dt_timestamp = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtTimestamp(
        expr=expr.Column(name="dt"),
    ),
    expected_output_records=[
        {"dt": 1623767445123000},
    ],
)
case_dt_total_days = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtTotalDays(
        expr=expr.Minus(
            left=expr.Column(name="dt"),
            right=datetime(1970, 1, 1),
        ),
    ),
    expected_output_records=[
        {"dt": 18793},
    ],
)
case_dt_total_hours = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtTotalHours(
        expr=expr.Minus(
            left=expr.Column(name="dt"),
            right=datetime(1970, 1, 1),
        ),
    ),
    expected_output_records=[
        {"dt": 451046},
    ],
)
case_dt_total_minutes = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtTotalMinutes(
        expr=expr.Minus(
            left=expr.Column(name="dt"),
            right=datetime(1970, 1, 1),
        ),
    ),
    expected_output_records=[
        {"dt": 27062790},
    ],
)
case_dt_total_seconds = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtTotalSeconds(
        expr=expr.Minus(
            left=expr.Column(name="dt"),
            right=datetime(1970, 1, 1),
        ),
    ),
    expected_output_records=[
        {"dt": 1623767445},
    ],
)
case_dt_total_milliseconds = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtTotalMilliSeconds(
        expr=expr.Minus(
            left=expr.Column(name="dt"),
            right=datetime(1970, 1, 1),
        ),
    ),
    expected_output_records=[
        {"dt": 1623767445123},
    ],
)
case_dt_total_microseconds = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtTotalMicroSeconds(
        expr=expr.Minus(
            left=expr.Column(name="dt"),
            right=datetime(1970, 1, 1),
        ),
    ),
    expected_output_records=[
        {"dt": 1623767445123000},
    ],
)
case_dt_total_nanoseconds = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtTotalNanoSeconds(
        expr=expr.Minus(
            left=expr.Column(name="dt"),
            right=datetime(1970, 1, 1),
        ),
    ),
    expected_output_records=[
        {"dt": 1623767445123000000},
    ],
)
case_dt_truncate_1 = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtTruncate(
        expr=expr.Column(name="dt"),
        every="1h",
    ),
    expected_output_records=[
        {"dt": datetime(2021, 6, 15, 14, 0, 0, 0)},
    ],
)
case_dt_truncate_2 = Case(
    input_records=[
        {"dt": datetime(2021, 6, 15, 14, 30, 45, 123000)},
    ],
    expr=expr.DtTruncate(
        expr=expr.Column(name="dt"),
        every=timedelta(hours=1),
    ),
    expected_output_records=[
        {"dt": datetime(2021, 6, 15, 14, 0, 0, 0)},
    ],
)
# fmt: on


def test():
    print("")

    case_datetime_to_string_1.run_with_columns_test()
    case_datetime_to_string_2.run_with_columns_test()
    case_dt_year.run_with_columns_test()
    case_dt_quarter.run_with_columns_test()
    case_dt_month.run_with_columns_test()
    case_dt_day.run_with_columns_test()
    case_dt_hour.run_with_columns_test()
    case_dt_minute.run_with_columns_test()
    case_dt_second.run_with_columns_test()
    case_dt_nanosecond.run_with_columns_test()
    case_dt_epoch.run_with_columns_test()
    case_dt_timestamp.run_with_columns_test()
    case_dt_total_days.run_with_columns_test()
    case_dt_total_hours.run_with_columns_test()
    case_dt_total_minutes.run_with_columns_test()
    case_dt_total_seconds.run_with_columns_test()
    case_dt_total_milliseconds.run_with_columns_test()
    case_dt_total_microseconds.run_with_columns_test()
    case_dt_total_nanoseconds.run_with_columns_test()
    case_dt_truncate_1.run_with_columns_test()
    case_dt_truncate_2.run_with_columns_test()


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.expr.dt", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl

from ..sentinel import NOTHING, REQUIRED, OPTIONAL
from ..expr import api as expr
from .. import utils_expr
from ..base_dfop import DfopEnum, BaseDfop, dfop_enum_to_klass_mapping

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_DFOP
    from ..expr.api import T_EXPR
    from ..typehint import IntoExpr, ColumnNameOrSelector


def _extract_exprs_named_exprs(exprs, named_exprs):
    """
    Used in Select.from_dict and WithColumns.from_dict.
    """
    new_exprs = [utils_expr.to_jsonpolars_into_expr(expr_like) for expr_like in exprs]
    new_named_exprs = {
        name: utils_expr.to_jsonpolars_into_expr(expr_like)
        for name, expr_like in named_exprs.items()
    }
    return new_exprs, new_named_exprs


def _convert_to_exprs_named_exprs(exprs, named_exprs):
    """
    Used in Select.to_polars and WithColumns.to_polars.
    """
    new_exprs = [utils_expr.to_polars_into_expr(expr_like) for expr_like in exprs]
    new_named_exprs = {
        name: utils_expr.to_polars_into_expr(expr_like)
        for name, expr_like in named_exprs.items()
    }
    return new_exprs, new_named_exprs


@dataclasses.dataclass
class Select(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.select.html
    """

    type: str = dataclasses.field(default=DfopEnum.select.value)
    exprs: T.List["IntoExpr"] = dataclasses.field(default_factory=list)
    named_exprs: T.Dict[str, "IntoExpr"] = dataclasses.field(default_factory=dict)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        exprs, named_exprs = _extract_exprs_named_exprs(
            dct["exprs"], dct["named_exprs"]
        )
        return cls(exprs=exprs, named_exprs=named_exprs)

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        exprs, named_exprs = _convert_to_exprs_named_exprs(self.exprs, self.named_exprs)
        return df.select(*exprs, **named_exprs)


dfop_enum_to_klass_mapping[DfopEnum.select.value] = Select


@dataclasses.dataclass
class Rename(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.rename.html
    """

    type: str = dataclasses.field(default=DfopEnum.rename.value)
    mapping: T.Union[T.Dict[str, str], T.Callable[[str], str]] = dataclasses.field(
        default=REQUIRED
    )

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            mapping=dct["mapping"],
        )

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.rename(self.mapping)


dfop_enum_to_klass_mapping[DfopEnum.rename.value] = Rename


@dataclasses.dataclass
class Drop(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.drop.html
    """

    type: str = dataclasses.field(default=DfopEnum.drop.value)
    columns: T.List["ColumnNameOrSelector"] = dataclasses.field(default=REQUIRED)
    strict: bool = dataclasses.field(default=True)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        columns, _ = _extract_exprs_named_exprs(dct["columns"], {})
        return cls(
            columns=columns,
            strict=dct["strict"],
        )

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        columns, _ = _convert_to_exprs_named_exprs(self.columns, {})
        return df.drop(*columns, strict=self.strict)


dfop_enum_to_klass_mapping[DfopEnum.drop.value] = Drop


@dataclasses.dataclass
class WithColumns(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.with_columns.html
    """

    type: str = dataclasses.field(default=DfopEnum.with_columns.value)
    exprs: T.List["IntoExpr"] = dataclasses.field(default_factory=list)
    named_exprs: T.Dict[str, "IntoExpr"] = dataclasses.field(default_factory=dict)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        exprs, named_exprs = _extract_exprs_named_exprs(
            dct["exprs"], dct["named_exprs"]
        )
        return cls(exprs=exprs, named_exprs=named_exprs)

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        exprs, named_exprs = _convert_to_exprs_named_exprs(self.exprs, self.named_exprs)
        return df.with_columns(*exprs, **named_exprs)


dfop_enum_to_klass_mapping[DfopEnum.with_columns.value] = WithColumns


@dataclasses.dataclass
class Head(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.head.html
    """

    type: str = dataclasses.field(default=DfopEnum.head.value)
    n: int = dataclasses.field(default=5)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(n=dct["n"])

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.head(self.n)


dfop_enum_to_klass_mapping[DfopEnum.head.value] = Head


@dataclasses.dataclass
class Tail(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.tail.html
    """

    type: str = dataclasses.field(default=DfopEnum.tail.value)
    n: int = dataclasses.field(default=5)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(n=dct["n"])

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.tail(self.n)


dfop_enum_to_klass_mapping[DfopEnum.tail.value] = Tail


@dataclasses.dataclass
class Sort(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.sort.html
    """

    type: str = dataclasses.field(default=DfopEnum.sort.value)
    by: T.List["IntoExpr"] = dataclasses.field(default=REQUIRED)
    descending: T.Union[bool, T.List[bool]] = dataclasses.field(default=False)
    nulls_last: T.Union[bool, T.List[bool]] = dataclasses.field(default=False)
    multithreaded: bool = dataclasses.field(default=True)
    maintain_order: bool = dataclasses.field(default=False)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            by=[
                utils_expr.to_jsonpolars_into_expr(expr_like) for expr_like in dct["by"]
            ],
            descending=dct["descending"],
            nulls_last=dct["nulls_last"],
            multithreaded=dct["multithreaded"],
            maintain_order=dct["maintain_order"],
        )

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.sort(
            *[utils_expr.to_polars_into_expr(expr_like) for expr_like in self.by],
            descending=self.descending,
            nulls_last=self.nulls_last,
            multithreaded=self.multithreaded,
            maintain_order=self.maintain_order,
        )


dfop_enum_to_klass_mapping[DfopEnum.sort.value] = Sort


@dataclasses.dataclass
class DropNulls(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.drop_nulls.html
    """

    type: str = dataclasses.field(default=DfopEnum.drop_nulls.value)
    subset: T.List["ColumnNameOrSelector"] = dataclasses.field(default=None)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        if dct["subset"] is None:
            subset = None
        else:
            subset, _ = _extract_exprs_named_exprs(dct["subset"], {})
        return cls(subset=subset)

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.drop_nulls(subset=self.subset)


dfop_enum_to_klass_mapping[DfopEnum.drop_nulls.value] = DropNulls



# -*- coding: utf-8 -*-

from jsonpolars import api


def test():
    _ = api
    _ = api.ExprEnum
    _ = api.parse_expr
    _ = api.T_EXPR
    _ = api.expr
    _ = api.DfopEnum
    _ = api.parse_dfop
    _ = api.T_DFOP
    _ = api.dfop

    # --- expr ---
    _ = api.expr.Cast
    _ = api.expr.List
    _ = api.expr.ListGet
    _ = api.expr.Datetime
    _ = api.expr.DtToString
    _ = api.expr.DtYear
    _ = api.expr.DtQuarter
    _ = api.expr.DtMonth
    _ = api.expr.DtDay
    _ = api.expr.DtHour
    _ = api.expr.DtMinute
    _ = api.expr.DtSecond
    _ = api.expr.DtNanoSecond
    _ = api.expr.DtEpoch
    _ = api.expr.DtTimestamp
    _ = api.expr.DtTotalDays
    _ = api.expr.DtTotalHours
    _ = api.expr.DtTotalMinutes
    _ = api.expr.DtTotalSeconds
    _ = api.expr.DtTotalMilliSeconds
    _ = api.expr.DtTotalMicroSeconds
    _ = api.expr.DtTotalNanoSeconds
    _ = api.expr.DtTruncate
    _ = api.expr.Plus
    _ = api.expr.Minus
    _ = api.expr.Multiply
    _ = api.expr.TrueDiv
    _ = api.expr.FloorDiv
    _ = api.expr.Negative
    _ = api.expr.Pow
    _ = api.expr.Equal
    _ = api.expr.NotEqual
    _ = api.expr.GreatThan
    _ = api.expr.GreatThanOrEqual
    _ = api.expr.LessThan
    _ = api.expr.LessThanOrEqual
    _ = api.expr.LogicalAnd
    _ = api.expr.LogicalOr
    _ = api.expr.String
    _ = api.expr.Split
    _ = api.expr.StrJoin
    _ = api.expr.StrContains
    _ = api.expr.StrDecode
    _ = api.expr.StrEncode
    _ = api.expr.StrStartsWith
    _ = api.expr.StrEndsWith
    _ = api.expr.Lit
    _ = api.expr.Column
    _ = api.expr.Alias

    # --- dfop ---
    _ = api.dfop.Select
    _ = api.dfop.Rename
    _ = api.dfop.Drop
    _ = api.dfop.WithColumns
    _ = api.dfop.Head
    _ = api.dfop.Tail
    _ = api.dfop.Sort
    _ = api.dfop.DropNulls
    _ = api.dfop.Count


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.api", preview=False)



# -*- coding: utf-8 -*-

import polars as pl
from jsonpolars.expr import api as expr
from jsonpolars.utils_expr import (
    to_jsonpolars_into_expr,
    to_polars_into_expr,
    parse_other_expr,
    other_expr_to_polars,
)


def test_to_jsonpolars_into_expr():
    assert to_jsonpolars_into_expr("col_1") == "col_1"
    assert to_jsonpolars_into_expr({"type": "column", "name": "col_1"}) == expr.Column(
        name="col_1"
    )
    assert to_jsonpolars_into_expr(expr.Column(name="col_1")) == expr.Column(
        name="col_1"
    )


def test_to_polars_into_expr():
    assert to_polars_into_expr("col_1") == "col_1"
    assert isinstance(to_polars_into_expr(expr.Column(name="col_1")), pl.Expr)


def test_parse_other_expr():
    assert parse_other_expr("hello") == "hello"
    assert parse_other_expr({"type": "column", "name": "col_1"}) == expr.Column(
        name="col_1"
    )


def test_other_expr_to_polars():
    assert other_expr_to_polars("hello") == "hello"
    assert isinstance(other_expr_to_polars(expr.Column(name="col_1")), pl.Expr)


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.utils_expr", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl

from ..sentinel import NOTHING, REQUIRED, OPTIONAL
from ..base_expr import ExprEnum, BaseExpr, expr_enum_to_klass_mapping, parse_expr

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_EXPR


@dataclasses.dataclass
class Cast(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.cast.html
    """
    type: str = dataclasses.field(default=ExprEnum.cast.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)
    dtype: pl.DataType = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            expr=parse_expr(dct["expr"]),
            dtype=dct["dtype"],
        )

    def to_polars(self) -> pl.Expr:
        return self.expr.to_polars().cast(self.dtype)


expr_enum_to_klass_mapping[ExprEnum.cast.value] = Cast



# -*- coding: utf-8 -*-

from jsonpolars.expr import api as expr
from jsonpolars.dfop import api as dfop
from jsonpolars.tests.dfop_case import Case


case_count = Case(
    input_records=[
        {"id": 1},
        {"id": 2},
        {"id": 3},
    ],
    dfop=dfop.Count(),
    expected_output_records=[
        {"id": 3},
    ],
)


def test():
    print("")

    case_count.run_test()


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.dfop.aggregation", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl

from ..sentinel import NOTHING, REQUIRED, OPTIONAL
from ..base_expr import ExprEnum, BaseExpr, expr_enum_to_klass_mapping, parse_expr
from ..utils_expr import parse_other_expr

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_EXPR
    from ..typehint import OtherExpr


def _other_expr_to_polars(other_expr: "OtherExpr"):
    """
    Convert jsonpolars expression to polars expression.

    Example

    >>> _other_expr_to_polars(1)
    1
    >>> _other_expr_to_polars("hello")
    'hello'
    >>> _other_expr_to_polars(Column("col_1"))
    pl.col("col_1")
    """
    if isinstance(other_expr, BaseExpr):
        return other_expr.to_polars()
    else:
        return other_expr


@dataclasses.dataclass
class Plus(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.add.html
    """

    type: str = dataclasses.field(default=ExprEnum.plus.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) + _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.plus.value] = Plus


@dataclasses.dataclass
class Minus(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.sub.html
    """

    type: str = dataclasses.field(default=ExprEnum.minus.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) - _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.minus.value] = Minus


@dataclasses.dataclass
class Multiply(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.mul.html
    """

    type: str = dataclasses.field(default=ExprEnum.mul.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) * _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.mul.value] = Multiply


@dataclasses.dataclass
class TrueDiv(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.truediv.html
    """

    type: str = dataclasses.field(default=ExprEnum.truediv.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) / _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.truediv.value] = TrueDiv


@dataclasses.dataclass
class FloorDiv(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.floordiv.html
    """

    type: str = dataclasses.field(default=ExprEnum.floordiv.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) // _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.floordiv.value] = FloorDiv


@dataclasses.dataclass
class Negative(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.neg.html
    """

    type: str = dataclasses.field(default=ExprEnum.neg.value)
    expr: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            expr=parse_other_expr(dct["expr"]),
        )

    def to_polars(self) -> pl.Expr:
        return -_other_expr_to_polars(self.expr)


expr_enum_to_klass_mapping[ExprEnum.neg.value] = Negative


@dataclasses.dataclass
class Pow(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.pow.html
    """

    type: str = dataclasses.field(default=ExprEnum.pow.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) ** _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.pow.value] = Pow


@dataclasses.dataclass
class Equal(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.eq.html
    """

    type: str = dataclasses.field(default=ExprEnum.eq.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) == _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.eq.value] = Equal


@dataclasses.dataclass
class NotEqual(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.ne.html
    """

    type: str = dataclasses.field(default=ExprEnum.ne.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) != _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.ne.value] = NotEqual


@dataclasses.dataclass
class GreatThan(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.gt.html
    """

    type: str = dataclasses.field(default=ExprEnum.gt.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) > _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.gt.value] = GreatThan


@dataclasses.dataclass
class GreatThanOrEqual(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.ge.html
    """

    type: str = dataclasses.field(default=ExprEnum.ge.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) >= _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.ge.value] = GreatThanOrEqual


@dataclasses.dataclass
class LessThan(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.lt.html
    """

    type: str = dataclasses.field(default=ExprEnum.lt.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) < _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.lt.value] = LessThan


@dataclasses.dataclass
class LessThanOrEqual(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.le.html
    """

    type: str = dataclasses.field(default=ExprEnum.le.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) <= _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.le.value] = LessThanOrEqual


@dataclasses.dataclass
class LogicalAnd(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.and_.html
    """

    type: str = dataclasses.field(default=ExprEnum.and_.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) & _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.and_.value] = LogicalAnd


@dataclasses.dataclass
class LogicalOr(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.or_.html
    """

    type: str = dataclasses.field(default=ExprEnum.or_.value)
    left: "OtherExpr" = dataclasses.field(default=REQUIRED)
    right: "OtherExpr" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            left=parse_other_expr(dct["left"]),
            right=parse_other_expr(dct["right"]),
        )

    def to_polars(self) -> pl.Expr:
        return _other_expr_to_polars(self.left) | _other_expr_to_polars(self.right)


expr_enum_to_klass_mapping[ExprEnum.or_.value] = LogicalOr



# -*- coding: utf-8 -*-

import typing as T
import polars as pl

from .base_expr import BaseExpr, parse_expr
from .base_dfop import BaseDfop, parse_dfop
from .typehint import IntoExpr


if T.TYPE_CHECKING:  # pragma: no cover
    from .expr.api import T_EXPR
    from .dfop.api import T_DFOP



# -*- coding: utf-8 -*-

import polars as pl

from jsonpolars.expr import api as expr
from jsonpolars.tests.expr_case import Case


case_list_get_1 = Case(
    input_records=[
        {"lst": [1, 2]},
    ],
    expr=expr.ListGet(
        expr=expr.Column(name="lst"),
        index=0,
        # index=expr.Lit(value=1),
    ),
    expected_output_records=[
        {"lst": 1},
    ],
)

case_list_get_2 = Case(
    input_records=[
        {"lst": [1, 2]},
    ],
    expr=expr.ListGet(
        expr=expr.List(expr=expr.Column(name="lst")),
        index=expr.Lit(value=1),
    ),
    expected_output_records=[
        {"lst": 2},
    ],
)


def test():
    print("")

    case_list_get_1.run_with_columns_test()
    case_list_get_2.run_with_columns_test()


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.expr.list", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl

from ..sentinel import NOTHING, REQUIRED, OPTIONAL
from ..base_expr import ExprEnum, BaseExpr, expr_enum_to_klass_mapping, parse_expr
from ..utils_expr import parse_other_expr, other_expr_to_polars

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_EXPR


def ensure_string(expr: "T_EXPR") -> pl.Expr:
    if isinstance(expr, String):
        return expr.to_polars()
    else:
        return expr.to_polars().str


@dataclasses.dataclass
class String(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/string.html
    """

    type: str = dataclasses.field(default=ExprEnum.string.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            expr=parse_expr(dct["expr"]),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr)


expr_enum_to_klass_mapping[ExprEnum.string.value] = String


@dataclasses.dataclass
class Split(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.split.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_split.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)
    by: str = dataclasses.field(default=REQUIRED)
    inclusive: bool = dataclasses.field(default=False)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            expr=parse_expr(dct["expr"]),
            by=dct["by"],
            inclusive=dct["inclusive"],
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).split(by=self.by, inclusive=self.inclusive)


expr_enum_to_klass_mapping[ExprEnum.str_split.value] = Split


@dataclasses.dataclass
class StrJoin(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.join.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_join.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)
    delimiter: str = dataclasses.field(default="")
    ignore_nulls: bool = dataclasses.field(default=True)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            expr=parse_expr(dct["expr"]),
            delimiter=dct.get("delimiter", ""),
            ignore_nulls=dct.get("ignore_nulls", True),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).join(
            delimiter=self.delimiter,
            ignore_nulls=self.ignore_nulls,
        )


expr_enum_to_klass_mapping[ExprEnum.str_join.value] = StrJoin


@dataclasses.dataclass
class StrContains(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.contains.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_contains.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)
    pattern: T.Union[str, "T_EXPR"] = dataclasses.field(default=REQUIRED)
    literal: bool = dataclasses.field(default=False)
    strict: bool = dataclasses.field(default=True)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            expr=parse_expr(dct["expr"]),
            pattern=parse_other_expr(dct["pattern"]),
            literal=dct["literal"],
            strict=dct["strict"],
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).contains(
            pattern=other_expr_to_polars(self.pattern),
            literal=self.literal,
            strict=self.strict,
        )


expr_enum_to_klass_mapping[ExprEnum.str_contains.value] = StrContains


@dataclasses.dataclass
class StrDecode(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.decode.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_decode.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)
    encoding: str = dataclasses.field(default=REQUIRED)
    strict: bool = dataclasses.field(default=True)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            expr=parse_expr(dct["expr"]),
            encoding=dct["encoding"],
            strict=dct["strict"],
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).decode(
            encoding=self.encoding,
            strict=self.strict,
        )


expr_enum_to_klass_mapping[ExprEnum.str_decode.value] = StrDecode


@dataclasses.dataclass
class StrEncode(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.encode.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_encode.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)
    encoding: str = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            expr=parse_expr(dct["expr"]),
            encoding=dct["encoding"],
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).encode(encoding=self.encoding)


expr_enum_to_klass_mapping[ExprEnum.str_encode.value] = StrEncode


@dataclasses.dataclass
class StrStartsWith(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.starts_with.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_starts_with.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)
    prefix: T.Union[str, "T_EXPR"] = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            expr=parse_expr(dct["expr"]),
            prefix=parse_other_expr(dct["prefix"]),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).starts_with(
            prefix=other_expr_to_polars(self.prefix)
        )


expr_enum_to_klass_mapping[ExprEnum.str_starts_with.value] = StrStartsWith


@dataclasses.dataclass
class StrEndsWith(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.ends_with.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_ends_with.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)
    suffix: T.Union[str, "T_EXPR"] = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            expr=parse_expr(dct["expr"]),
            suffix=parse_other_expr(dct["suffix"]),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).ends_with(
            suffix=other_expr_to_polars(self.suffix)
        )


expr_enum_to_klass_mapping[ExprEnum.str_ends_with.value] = StrEndsWith



# -*- coding: utf-8 -*-

from datetime import datetime

import polars as pl

from jsonpolars.expr import api as expr
from jsonpolars.tests.expr_case import Case

case_cast = Case(
    input_records=[
        {"time": "2024-08-01T12:30:45"},
    ],
    expr=expr.Cast(
        dtype=pl.Datetime(),
        expr=expr.Column(name="time"),
    ),
    expected_output_records=[
        {"time": datetime(2024, 8, 1, 12, 30, 45)},
    ],
)


def test():
    print("")

    case_cast.run_with_columns_test()


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.expr.manipulation", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import sys


def make_sentinel(name="_MISSING", var_name=None):  # pragma: no cover
    """Creates and returns a new **instance** of a new class, suitable for
    usage as a "sentinel", a kind of singleton often used to indicate
    a value is missing when ``None`` is a valid input.
    Args:
        name (str): Name of the Sentinel
        var_name (str): Set this name to the name of the variable in
            its respective module enable pickleability. Note:
            pickleable sentinels should be global constants at the top
            level of their module.
    >>> make_sentinel(var_name='_MISSING')
    _MISSING
    The most common use cases here in boltons are as default values
    for optional function arguments, partly because of its
    less-confusing appearance in automatically generated
    documentation. Sentinels also function well as placeholders in queues
    and linked lists.
    .. note::
      By design, additional calls to ``make_sentinel`` with the same
      values will not produce equivalent objects.
      >>> make_sentinel('TEST') == make_sentinel('TEST')
      False
      >>> type(make_sentinel('TEST')) == type(make_sentinel('TEST'))
      False
    """

    class Sentinel(object):
        def __init__(self):
            self.name = name
            self.var_name = var_name

        def __repr__(self):
            if self.var_name:
                return self.var_name
            return "%s(%r)" % (self.__class__.__name__, self.name)

        if var_name:

            def __reduce__(self):
                return self.var_name

        def __nonzero__(self):
            return False

        __bool__ = __nonzero__

    if var_name:
        frame = sys._getframe(1)
        module = frame.f_globals.get("__name__")
        if not module or module not in sys.modules:
            raise ValueError(
                "Pickleable sentinel objects (with var_name) can only"
                " be created from top-level module scopes"
            )
        Sentinel.__module__ = module

    return Sentinel()


NOTHING = make_sentinel(name="NOTHING")
REQUIRED = make_sentinel(name="REQUIRED")
OPTIONAL = make_sentinel(name="OPTIONAL")


def resolve_kwargs(
    _mapper: T.Optional[T.Dict[str, str]] = None,
    **kwargs,
) -> dict:
    if _mapper is None:  # pragma: no cover
        _mapper = dict()
    return {
        _mapper.get(key, key): value
        for key, value in kwargs.items()
        if value is not NOTHING
    }



# -*- coding: utf-8 -*-

import typing as T
from .manipulation import Cast
from .list import List
from .list import ListGet
from .dt import Datetime
from .dt import DtToString
from .dt import DtYear
from .dt import DtQuarter
from .dt import DtMonth
from .dt import DtDay
from .dt import DtHour
from .dt import DtMinute
from .dt import DtSecond
from .dt import DtNanoSecond
from .dt import DtEpoch
from .dt import DtTimestamp
from .dt import DtTotalDays
from .dt import DtTotalHours
from .dt import DtTotalMinutes
from .dt import DtTotalSeconds
from .dt import DtTotalMilliSeconds
from .dt import DtTotalMicroSeconds
from .dt import DtTotalNanoSeconds
from .dt import DtTruncate
from .operator import Plus
from .operator import Minus
from .operator import Multiply
from .operator import TrueDiv
from .operator import FloorDiv
from .operator import Negative
from .operator import Pow
from .operator import Equal
from .operator import NotEqual
from .operator import GreatThan
from .operator import GreatThanOrEqual
from .operator import LessThan
from .operator import LessThanOrEqual
from .operator import LogicalAnd
from .operator import LogicalOr
from .string import String
from .string import Split
from .string import StrJoin
from .string import StrContains
from .string import StrDecode
from .string import StrEncode
from .string import StrStartsWith
from .string import StrEndsWith
from .function import Lit
from .column import Column
from .column import Alias

T_EXPR = T.Union[
    Cast,
    List,
    ListGet,
    Datetime,
    DtToString,
    DtYear,
    DtQuarter,
    DtMonth,
    DtDay,
    DtHour,
    DtMinute,
    DtSecond,
    DtNanoSecond,
    DtEpoch,
    DtTimestamp,
    DtTotalDays,
    DtTotalHours,
    DtTotalMinutes,
    DtTotalSeconds,
    DtTotalMilliSeconds,
    DtTotalMicroSeconds,
    DtTotalNanoSeconds,
    DtTruncate,
    Plus,
    Minus,
    Multiply,
    TrueDiv,
    FloorDiv,
    Negative,
    Pow,
    Equal,
    NotEqual,
    GreatThan,
    GreatThanOrEqual,
    LessThan,
    LessThanOrEqual,
    LogicalAnd,
    LogicalOr,
    String,
    Split,
    StrJoin,
    StrContains,
    StrDecode,
    StrEncode,
    StrStartsWith,
    StrEndsWith,
    Lit,
    Column,
    Alias,
]



# -*- coding: utf-8 -*-

"""
Package Description.
"""

from ._version import __version__

__short_description__ = "Package short description."
__license__ = "MIT"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__maintainer__ = "Sanhe Hu"
__maintainer_email__ = "husanhe@email.com"
__github_username__ = "MacHu-GWU"



# -*- coding: utf-8 -*-

from jsonpolars.expr import api as expr
from jsonpolars.dfop import api as dfop
from jsonpolars.tests.dfop_case import Case


case_select = Case(
    input_records=[
        {"a": 1, "b": 2, "c": 3},
    ],
    dfop=dfop.Select(
        exprs=[
            "a",
            expr.Column(name="c"),
        ],
        named_exprs={
            "d": "b",
            "e": expr.Lit(value=5),
        },
    ),
    expected_output_records=[
        {"a": 1, "c": 3, "d": 2, "e": 5},
    ],
)
case_rename = Case(
    input_records=[
        {"a": 1, "b": 2, "c": 3},
    ],
    dfop=dfop.Rename(
        mapping={
            "a": "x",
            "c": "z",
        }
    ),
    expected_output_records=[
        {"x": 1, "b": 2, "z": 3},
    ],
)
case_drop = Case(
    input_records=[
        {"a": 1, "b": 2, "c": 3},
    ],
    dfop=dfop.Drop(columns=["a", expr.Column(name="c")]),
    expected_output_records=[
        {"b": 2},
    ],
)
case_with_columns = Case(
    input_records=[
        {"a": 1},
    ],
    dfop=dfop.WithColumns(
        exprs=[
            expr.Alias(
                name="b",
                expr=expr.Plus(
                    left=expr.Column(name="a"),
                    right=expr.Lit(value=1),
                ),
            ),
        ],
        named_exprs={
            "c": expr.Plus(
                left=expr.Column(name="a"),
                right=expr.Lit(value=2),
            )
        },
    ),
    expected_output_records=[
        {"a": 1, "b": 2, "c": 3},
    ],
)
case_head = Case(
    input_records=[
        {"id": 1},
        {"id": 2},
        {"id": 3},
        {"id": 4},
    ],
    dfop=dfop.Head(n=2),
    expected_output_records=[
        {"id": 1},
        {"id": 2},
    ],
)
case_tail = Case(
    input_records=[
        {"id": 1},
        {"id": 2},
        {"id": 3},
        {"id": 4},
    ],
    dfop=dfop.Tail(n=2),
    expected_output_records=[
        {"id": 3},
        {"id": 4},
    ],
)
case_count = Case(
    input_records=[
        {"id": 1},
        {"id": 2},
        {"id": 3},
    ],
    dfop=dfop.Count(),
    expected_output_records=[
        {"id": 3},
    ],
)
case_sort_1 = Case(
    input_records=[
        {"id": 2},
        {"id": 3},
        {"id": 1},
    ],
    dfop=dfop.Sort(by=["id"]),
    expected_output_records=[
        {"id": 1},
        {"id": 2},
        {"id": 3},
    ],
)
case_sort_2 = Case(
    input_records=[
        {"id": 1, "name": "a"},
        {"id": 1, "name": "b"},
        {"id": 2, "name": "c"},
        {"id": 2, "name": "d"},
    ],
    dfop=dfop.Sort(by=["id", "name"], descending=[True, False]),
    expected_output_records=[
        {"id": 2, "name": "c"},
        {"id": 2, "name": "d"},
        {"id": 1, "name": "a"},
        {"id": 1, "name": "b"},
    ],
)
case_drop_nulls_1 = Case(
    input_records=[
        {"id": 1, "name": "a"},
        {"id": 2, "name": None},
        {"id": None, "name": "c"},
    ],
    dfop=dfop.DropNulls(),
    expected_output_records=[
        {"id": 1, "name": "a"},
    ],
)
case_drop_nulls_2 = Case(
    input_records=[
        {"id": 1, "name": "a"},
        {"id": 2, "name": None},
        {"id": None, "name": "c"},
    ],
    dfop=dfop.DropNulls(subset=["id"]),
    expected_output_records=[
        {"id": 1, "name": "a"},
        {"id": 2, "name": None},
    ],
)
case_drop_nulls_3 = Case(
    input_records=[
        {"id": 1, "name": "a"},
        {"id": 2, "name": None},
        {"id": None, "name": "c"},
    ],
    dfop=dfop.DropNulls(subset=["name"]),
    expected_output_records=[
        {"id": 1, "name": "a"},
        {"id": None, "name": "c"},
    ],
)


def test():
    print("")

    case_select.run_test()
    case_rename.run_test()
    case_drop.run_test()
    case_with_columns.run_test()
    case_head.run_test()
    case_tail.run_test()
    case_count.run_test()
    case_sort_1.run_test()
    case_sort_2.run_test()
    case_drop_nulls_1.run_test()
    case_drop_nulls_2.run_test()
    case_drop_nulls_3.run_test()


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.dfop.manipulation", preview=False)



# -*- coding: utf-8 -*-

import polars as pl

from jsonpolars.expr import api as expr
from jsonpolars.tests.expr_case import Case


case_split_1 = Case(
    input_records=[
        {"id": "a-1"},
    ],
    expr=expr.Split(
        expr=expr.Column(name="id"),
        by="-",
    ),
    expected_output_records=[
        {"id": ["a", "1"]},
    ],
)

case_split_2 = Case(
    input_records=[
        {"id": "a-1"},
    ],
    expr=expr.Split(
        expr=expr.String(expr=expr.Column(name="id")),
        by="-",
    ),
    expected_output_records=[
        {"id": ["a", "1"]},
    ],
)
case_join = Case(
    input_records=[
        {"foo": "1"},
        {"foo": "2"},
        {"foo": "3"},
    ],
    expr=expr.StrJoin(
        delimiter="-",
        expr=expr.Column(name="foo"),
    ),
    expected_output_records=[
        {"foo": "1-2-3"},
        {"foo": "1-2-3"},
        {"foo": "1-2-3"},
    ],
)
case_contains_1 = Case(
    input_records=[
        {"foo": "123-456-7890"},
    ],
    expr=expr.StrContains(
        expr=expr.Column(name="foo"),
        pattern="456",
    ),
    expected_output_records=[
        {"foo": True},
    ],
)
case_contains_2 = Case(
    input_records=[
        {"foo": "123-456-7890"},
    ],
    expr=expr.StrContains(
        expr=expr.Column(name="foo"),
        pattern=r"\d{4}",
    ),
    expected_output_records=[
        {"foo": True},
    ],
)
case_contains_3 = Case(
    input_records=[
        {"foo": "123-456-7890"},
    ],
    expr=expr.StrContains(
        expr=expr.Column(name="foo"),
        pattern=expr.Lit(value=r"\d{10}"),
    ),
    expected_output_records=[
        {"foo": False},
    ],
)
case_decode = Case(
    input_records=[
        {"color": "ffff00"},
    ],
    expr=expr.StrDecode(
        expr=expr.Column(name="color"),
        encoding="hex",
    ),
    expected_output_records=[
        {"color": b"\xff\xff\x00"},
    ],
)
case_encode = Case(
    input_records=[
        {"word": "foo"},
    ],
    expr=expr.StrEncode(
        expr=expr.Column(name="word"),
        encoding="hex",
    ),
    expected_output_records=[
        {"word": "666f6f"},
    ],
)
case_starts_with_1 = Case(
    input_records=[
        {"foo": "Alice"},
    ],
    expr=expr.StrStartsWith(
        expr=expr.Column(name="foo"),
        prefix="A",
    ),
    expected_output_records=[
        {"foo": True},
    ],
)
case_starts_with_2 = Case(
    input_records=[
        {"foo": "Alice"},
    ],
    expr=expr.StrStartsWith(
        expr=expr.Column(name="foo"),
        prefix=expr.Lit(value="B"),
    ),
    expected_output_records=[
        {"foo": False},
    ],
)
case_starts_with_3 = Case(
    input_records=[
        {"foo": "Alice", "prefix": "A"},
    ],
    expr=expr.StrStartsWith(
        expr=expr.Column(name="foo"),
        prefix=expr.Column(name="prefix"),
    ),
    expected_output_records=[
        {"foo": True, "prefix": "A"},
    ],
)
case_ends_with_1 = Case(
    input_records=[
        {"foo": "Alice"},
    ],
    expr=expr.StrEndsWith(
        expr=expr.Column(name="foo"),
        suffix="e",
    ),
    expected_output_records=[
        {"foo": True},
    ],
)
case_ends_with_2 = Case(
    input_records=[
        {"foo": "Alice"},
    ],
    expr=expr.StrEndsWith(
        expr=expr.Column(name="foo"),
        suffix=expr.Lit(value="f"),
    ),
    expected_output_records=[
        {"foo": False},
    ],
)
case_ends_with_3 = Case(
    input_records=[
        {"foo": "Alice", "suffix": "e"},
    ],
    expr=expr.StrEndsWith(
        expr=expr.Column(name="foo"),
        suffix=expr.Column(name="suffix"),
    ),
    expected_output_records=[
        {"foo": True, "suffix": "e"},
    ],
)


def test():
    print("")

    case_split_1.run_with_columns_test()
    case_split_2.run_with_columns_test()
    case_join.run_with_columns_test()
    case_contains_1.run_with_columns_test()
    case_contains_2.run_with_columns_test()
    case_contains_3.run_with_columns_test()
    case_decode.run_with_columns_test()
    case_encode.run_with_columns_test()
    case_starts_with_1.run_with_columns_test()
    case_starts_with_2.run_with_columns_test()
    case_starts_with_3.run_with_columns_test()
    case_ends_with_1.run_with_columns_test()
    case_ends_with_2.run_with_columns_test()
    case_ends_with_3.run_with_columns_test()


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.expr.string", preview=False)



# -*- coding: utf-8 -*-

from jsonpolars.expr import api as expr
from jsonpolars.tests.expr_case import Case

case_alias = Case(
    input_records=[
        {"a": 1},
    ],
    expr=expr.Alias(
        name="b",
        expr=expr.Column(name="a"),
    ),
    expected_output_records=[
        {"a": 1, "b": 1},
    ],
)


def test():
    print("")

    case_alias.run_with_columns_test()


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.expr.column", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import polars as pl

from .base_expr import BaseExpr, parse_expr
from .typehint import IntoExpr, PythonLiteral, OtherExpr


if T.TYPE_CHECKING:  # pragma: no cover
    from .expr.api import T_EXPR


def to_jsonpolars_into_expr(expr_like: T.Union[str, dict, "T_EXPR"]) -> IntoExpr:
    if isinstance(expr_like, str):
        return expr_like
    elif isinstance(expr_like, dict):
        return parse_expr(expr_like)
    elif isinstance(expr_like, BaseExpr):
        return expr_like
    else:  # pragma: no cover
        raise NotImplementedError(f"Unsupported type: {type(expr_like)}")


def to_polars_into_expr(expr_like: T.Union[T.Any, "T_EXPR"]):
    if isinstance(expr_like, BaseExpr):
        return expr_like.to_polars()
    else:  # pragma: no cover
        return expr_like


def parse_other_expr(value) -> "OtherExpr":
    if isinstance(value, dict):
        return parse_expr(value)
    else:
        return value


def other_expr_to_polars(expr: "OtherExpr") -> T.Union["PythonLiteral", "pl.Expr"]:
    if isinstance(expr, BaseExpr):
        return expr.to_polars()
    else:
        return expr



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl

from ..sentinel import NOTHING, REQUIRED, OPTIONAL
from ..base_expr import ExprEnum, BaseExpr, expr_enum_to_klass_mapping, parse_expr

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_EXPR


@dataclasses.dataclass
class Lit(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.lit.html
    """

    type: str = dataclasses.field(default=ExprEnum.lit.value)
    value: T.Any = dataclasses.field(default=REQUIRED)
    dtype: T.Optional["pl.DataType"] = dataclasses.field(default=None)
    allow_object: bool = dataclasses.field(default=False)

    def to_polars(self) -> pl.Expr:
        return pl.lit(
            value=self.value,
            dtype=self.dtype,
            allow_object=self.allow_object,
        )


expr_enum_to_klass_mapping[ExprEnum.lit.value] = Lit



# -*- coding: utf-8 -*-

import typing as T
from .manipulation import Select
from .manipulation import Rename
from .manipulation import Drop
from .manipulation import WithColumns
from .manipulation import Head
from .manipulation import Tail
from .manipulation import Sort
from .manipulation import DropNulls
from .aggregation import Count

T_DFOP = T.Union[
    Select,
    Rename,
    Drop,
    WithColumns,
    Head,
    Tail,
    Sort,
    DropNulls,
    Count,
]



# -*- coding: utf-8 -*-

import typing as T
import dataclasses
from datetime import timedelta

import polars as pl

from ..sentinel import NOTHING, REQUIRED, OPTIONAL
from ..base_expr import ExprEnum, BaseExpr, expr_enum_to_klass_mapping, parse_expr
from ..utils_expr import parse_other_expr


if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_EXPR


def ensure_datetime(expr: "T_EXPR") -> pl.Expr:
    if isinstance(expr, Datetime):
        return expr.to_polars()
    else:
        return expr.to_polars().dt


@dataclasses.dataclass
class Datetime(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/temporal.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr)


expr_enum_to_klass_mapping[ExprEnum.dt.value] = Datetime


@dataclasses.dataclass
class DtToString(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.to_string.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_to_string.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)
    format: str = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            expr=parse_expr(dct["expr"]),
            format=dct["format"],
        )

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).to_string(format=self.format)


expr_enum_to_klass_mapping[ExprEnum.dt_to_string.value] = DtToString


@dataclasses.dataclass
class DtYear(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.year.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_year.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).year()


expr_enum_to_klass_mapping[ExprEnum.dt_year.value] = DtYear


@dataclasses.dataclass
class DtQuarter(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.quarter.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_quarter.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).quarter()


expr_enum_to_klass_mapping[ExprEnum.dt_quarter.value] = DtQuarter


@dataclasses.dataclass
class DtMonth(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.month.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_month.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).month()


expr_enum_to_klass_mapping[ExprEnum.dt_month.value] = DtMonth


@dataclasses.dataclass
class DtDay(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.day.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_day.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).day()


expr_enum_to_klass_mapping[ExprEnum.dt_day.value] = DtDay


@dataclasses.dataclass
class DtHour(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.hour.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_hour.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).hour()


expr_enum_to_klass_mapping[ExprEnum.dt_hour.value] = DtHour


@dataclasses.dataclass
class DtMinute(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.minute.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_minute.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).minute()


expr_enum_to_klass_mapping[ExprEnum.dt_minute.value] = DtMinute


@dataclasses.dataclass
class DtSecond(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.second.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_second.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).second()


expr_enum_to_klass_mapping[ExprEnum.dt_second.value] = DtSecond


@dataclasses.dataclass
class DtNanoSecond(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.nanosecond.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_nanosecond.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).nanosecond()


expr_enum_to_klass_mapping[ExprEnum.dt_nanosecond.value] = DtNanoSecond


@dataclasses.dataclass
class DtEpoch(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.epoch.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_epoch.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)
    time_unit: str = dataclasses.field(default="us")

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]), time_unit=dct["time_unit"])

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).epoch(time_unit=self.time_unit)


expr_enum_to_klass_mapping[ExprEnum.dt_epoch.value] = DtEpoch


@dataclasses.dataclass
class DtTimestamp(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.timestamp.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_timestamp.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)
    time_unit: str = dataclasses.field(default="us")

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]), time_unit=dct["time_unit"])

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).timestamp(time_unit=self.time_unit)


expr_enum_to_klass_mapping[ExprEnum.dt_timestamp.value] = DtTimestamp


@dataclasses.dataclass
class DtTotalDays(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.total_days.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_total_days.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).total_days()


expr_enum_to_klass_mapping[ExprEnum.dt_total_days.value] = DtTotalDays


@dataclasses.dataclass
class DtTotalHours(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.total_hours.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_total_hours.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).total_hours()


expr_enum_to_klass_mapping[ExprEnum.dt_total_hours.value] = DtTotalHours


@dataclasses.dataclass
class DtTotalMinutes(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.total_minutes.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_total_minutes.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).total_minutes()


expr_enum_to_klass_mapping[ExprEnum.dt_total_minutes.value] = DtTotalMinutes


@dataclasses.dataclass
class DtTotalSeconds(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.total_seconds.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_total_seconds.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).total_seconds()


expr_enum_to_klass_mapping[ExprEnum.dt_total_seconds.value] = DtTotalSeconds


@dataclasses.dataclass
class DtTotalMilliSeconds(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.total_milliseconds.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_total_milliseconds.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).total_milliseconds()


expr_enum_to_klass_mapping[ExprEnum.dt_total_milliseconds.value] = DtTotalMilliSeconds


@dataclasses.dataclass
class DtTotalMicroSeconds(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.total_microseconds.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_total_microseconds.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).total_microseconds()


expr_enum_to_klass_mapping[ExprEnum.dt_total_microseconds.value] = DtTotalMicroSeconds


@dataclasses.dataclass
class DtTotalNanoSeconds(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.total_nanoseconds.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_total_nanoseconds.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(expr=parse_expr(dct["expr"]))

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).total_nanoseconds()


expr_enum_to_klass_mapping[ExprEnum.dt_total_nanoseconds.value] = DtTotalNanoSeconds


@dataclasses.dataclass
class DtTruncate(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.dt.truncate.html
    """

    type: str = dataclasses.field(default=ExprEnum.dt_truncate.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)
    every: T.Union[str, timedelta, "T_EXPR"] = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            expr=parse_expr(dct["expr"]),
            every=parse_other_expr(dct["every"]),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_datetime(self.expr).truncate(every=self.every)


expr_enum_to_klass_mapping[ExprEnum.dt_truncate.value] = DtTruncate



# -*- coding: utf-8 -*-

import polars as pl

from jsonpolars.expr import api as expr
from jsonpolars.tests.expr_case import Case


def test():
    print("")


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.expr.function", preview=False)



# -*- coding: utf-8 -*-

import polars as pl

from jsonpolars.expr import api as expr
from jsonpolars.tests.expr_case import Case


case_plus_1 = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.Plus(
        left=expr.Column(name="id"),
        right=expr.Lit(value=100),
    ),
    expected_output_records=[
        {"id": 101},
    ],
)
case_plus_2 = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.Plus(
        left=expr.Column(name="id"),
        right=100,
    ),
    expected_output_records=[
        {"id": 101},
    ],
)
case_minus_1 = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.Minus(
        left=expr.Column(name="id"),
        right=expr.Lit(value=100),
    ),
    expected_output_records=[
        {"id": -99},
    ],
)
case_minus_2 = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.Minus(
        left=expr.Column(name="id"),
        right=100,
    ),
    expected_output_records=[
        {"id": -99},
    ],
)
case_multiply = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.Multiply(
        left=expr.Column(name="id"),
        right=5,
    ),
    expected_output_records=[
        {"id": 5},
    ],
)
case_truediv = Case(
    input_records=[
        {"id": 6},
    ],
    expr=expr.TrueDiv(
        left=expr.Column(name="id"),
        right=2,
    ),
    expected_output_records=[
        {"id": 3},
    ],
)
case_floordiv = Case(
    input_records=[
        {"id": 10},
    ],
    expr=expr.FloorDiv(
        left=expr.Column(name="id"),
        right=4,
    ),
    expected_output_records=[
        {"id": 2},
    ],
)
case_neg = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.Negative(
        expr=expr.Column(name="id"),
    ),
    expected_output_records=[
        {"id": -1},
    ],
)
case_pow = Case(
    input_records=[
        {"id": 2},
    ],
    expr=expr.Pow(
        left=expr.Column(name="id"),
        right=4,
    ),
    expected_output_records=[
        {"id": 16},
    ],
)
case_equal = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.Equal(
        left=expr.Column(name="id"),
        right=1,
    ),
    expected_output_records=[
        {"id": True},
    ],
)
case_notequal = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.NotEqual(
        left=expr.Column(name="id"),
        right=1,
    ),
    expected_output_records=[
        {"id": False},
    ],
)
case_gt_1 = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.GreatThan(
        left=expr.Column(name="id"),
        right=0,
    ),
    expected_output_records=[
        {"id": True},
    ],
)
case_gt_2 = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.GreatThan(
        left=expr.Column(name="id"),
        right=1,
    ),
    expected_output_records=[
        {"id": False},
    ],
)
case_ge_1 = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.GreatThanOrEqual(
        left=expr.Column(name="id"),
        right=0,
    ),
    expected_output_records=[
        {"id": True},
    ],
)
case_ge_2 = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.GreatThanOrEqual(
        left=expr.Column(name="id"),
        right=1,
    ),
    expected_output_records=[
        {"id": True},
    ],
)
case_lt_1 = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.LessThan(left=expr.Column(name="id"), right=2.0),
    expected_output_records=[
        {"id": True},
    ],
)
case_lt_2 = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.LessThan(
        left=expr.Column(name="id"),
        right=1,
    ),
    expected_output_records=[
        {"id": False},
    ],
)
case_le_1 = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.LessThanOrEqual(left=expr.Column(name="id"), right=2.0),
    expected_output_records=[
        {"id": True},
    ],
)
case_le_2 = Case(
    input_records=[
        {"id": 1},
    ],
    expr=expr.LessThanOrEqual(
        left=expr.Column(name="id"),
        right=1,
    ),
    expected_output_records=[
        {"id": True},
    ],
)
case_and_1 = Case(
    input_records=[
        {"flag": True},
    ],
    expr=expr.LogicalAnd(
        left=expr.Column(name="flag"),
        right=True,
    ),
    expected_output_records=[
        {"flag": True},
    ],
)
case_and_2 = Case(
    input_records=[
        {"flag": True},
    ],
    expr=expr.LogicalAnd(
        left=expr.Column(name="flag"),
        right=False,
    ),
    expected_output_records=[
        {"flag": False},
    ],
)
case_or_1 = Case(
    input_records=[
        {"flag": False},
    ],
    expr=expr.LogicalOr(
        left=expr.Column(name="flag"),
        right=False,
    ),
    expected_output_records=[
        {"flag": False},
    ],
)
case_or_2 = Case(
    input_records=[
        {"flag": False},
    ],
    expr=expr.LogicalOr(
        left=expr.Column(name="flag"),
        right=True,
    ),
    expected_output_records=[
        {"flag": True},
    ],
)


def test():
    print("")

    case_plus_1.run_with_columns_test()
    case_plus_2.run_with_columns_test()
    case_minus_1.run_with_columns_test()
    case_minus_2.run_with_columns_test()
    case_multiply.run_with_columns_test()
    case_truediv.run_with_columns_test()
    case_floordiv.run_with_columns_test()
    case_neg.run_with_columns_test()
    case_pow.run_with_columns_test()
    case_equal.run_with_columns_test()
    case_notequal.run_with_columns_test()
    case_gt_1.run_with_columns_test()
    case_gt_2.run_with_columns_test()
    case_ge_1.run_with_columns_test()
    case_ge_2.run_with_columns_test()
    case_lt_1.run_with_columns_test()
    case_lt_2.run_with_columns_test()
    case_le_1.run_with_columns_test()
    case_le_2.run_with_columns_test()
    case_and_1.run_with_columns_test()
    case_and_2.run_with_columns_test()
    case_or_1.run_with_columns_test()
    case_or_2.run_with_columns_test()


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.expr.operator", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl

from ..sentinel import NOTHING, REQUIRED, OPTIONAL
from ..expr import api as expr
from ..base_dfop import DfopEnum, BaseDfop, dfop_enum_to_klass_mapping

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_DFOP
    from ..expr.api import T_EXPR
    from ..typehint import IntoExpr, ColumnNameOrSelector


@dataclasses.dataclass
class Count(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.count.html
    """

    type: str = dataclasses.field(default=DfopEnum.count.value)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls()

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.count()


dfop_enum_to_klass_mapping[DfopEnum.count.value] = Count




# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl

from ..sentinel import NOTHING, REQUIRED, OPTIONAL
from ..base_expr import ExprEnum, BaseExpr, expr_enum_to_klass_mapping, parse_expr

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_EXPR


def ensure_list(expr: "T_EXPR") -> pl.Expr:
    if isinstance(expr, List):
        return expr.to_polars()
    else:
        return expr.to_polars().list


@dataclasses.dataclass
class List(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/list.html
    """
    type: str = dataclasses.field(default=ExprEnum.list.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            expr=parse_expr(dct["expr"]),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_list(self.expr)


expr_enum_to_klass_mapping[ExprEnum.list.value] = List


@dataclasses.dataclass
class ListGet(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.list.get.html#polars.Expr.list.get
    """
    type: str = dataclasses.field(default=ExprEnum.list_get.value)
    expr: "T_EXPR" = dataclasses.field(default=REQUIRED)
    index: T.Union[int, "T_EXPR"] = dataclasses.field(default=REQUIRED)
    null_on_oob: bool = dataclasses.field(default=False)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        index = dct["index"]
        if isinstance(index, int):
            pass
        elif isinstance(index, dict):
            index = parse_expr(index)
        else:  # pragma: no cover
            raise ValueError(f"Unknown index type: {index}")
        return cls(
            expr=parse_expr(dct["expr"]),
            index=index,
            null_on_oob=dct["null_on_oob"],
        )

    def to_polars(self) -> pl.Expr:
        expr = ensure_list(self.expr)
        if isinstance(self.index, int):
            index = self.index
        elif isinstance(self.index, BaseExpr):
            index = self.index.to_polars()
        return expr.get(index=index, null_on_oob=self.null_on_oob)


expr_enum_to_klass_mapping[ExprEnum.list_get.value] = ListGet
